"""Base Plotting module."""
from __future__ import annotations
from . import State, units
from CoolProp.CoolProp import PropsSI
import matplotlib.pyplot as plt
import numpy as np
from abc import ABC, abstractmethod
from dataclasses import dataclass, field


@dataclass
class PlottedState:
    """Data class to efficiently store states in the self.states dictionary."""

    key: str
    state: State
    # key: Plot axes string (Tv, pv)
    # value: Line2D instance for that plot of the marker for this state
    markers: dict = field(default_factory=dict)


class PlottingBase(ABC):
    """Basic Plotting manager for thermodynamic states.

    Parameters
    ----------
    substance : `str`
        One of the substances supported by CoolProp
    """

    axis_units = {
        "v": "m**3/kg",
        "T": "K",
        "s": "J/(kg*K)",
        "p": "pascal",
        "u": "J/kg",
        "h": "J/kg",
        "x": "dimensionless",
    }

    allowed_processes = {
        "isochoric": "v",
        "isovolumetric": "v",
        "isometric": "v",
        "isobaric": "p",
        "isothermal": "T",
        "isoenergetic": "u",
        "isoenthalpic": "h",
        "isentropic": "s",
    }

    def __init__(self, substance: str):
        self.states = {}
        self.plots = {}
        self.processes = {}

    @abstractmethod
    def plot(self, x_axis: str, y_axis: str):  # pragma: no cover
        """Hold the place of a plot function that a child class must establish."""
        pass

    def add_state(self, state: State, key: str | None = None, label: str | None = None):
        """Add a state to the self.states dictionary and plot it."""
        if key is None:
            key = repr(state)

        if label is not None:
            state.label = label

        plotted_state = PlottedState(key=key, state=state)

        for plot_key, value in self.plots.items():
            x_data = []
            y_data = []
            fig, axis = value
            x_axis, y_axis = plot_key
            x_data.append(getattr(state, x_axis).magnitude)
            y_data.append(getattr(state, y_axis).magnitude)
            x_data = np.array(x_data) * getattr(units, self.axis_units[x_axis])
            y_data = np.array(y_data) * getattr(units, self.axis_units[y_axis])
            (line,) = axis.plot(x_data, y_data, marker="o")
            if state.label is not None:
                axis.annotate(
                    state.label,
                    (x_data[0], y_data[0]),
                    textcoords="offset pixels",
                    xytext=(5, 5),
                )
            plotted_state.markers[plot_key] = line

        self.states[key] = plotted_state

    def remove_state(self, state: State | None = None, key: str | None = None):
        """Remove a state from the self.states dictionary and plots."""
        if state is None and key is None:
            raise ValueError("No state or key was entered. Unable to find state")
        if state is not None and repr(state) in self.states:
            state_to_be_removed = self.states[repr(state)]
        elif key is not None and key in self.states:
            state_to_be_removed = self.states[key]
        elif key is not None and key not in self.states and state is None:
            raise ValueError("Couldn't find key")
        else:
            for key, s_2 in self.states.items():
                if state == s_2.state:
                    state_to_be_removed = self.states[key]
                    break
            else:
                raise ValueError("Couldn't find the state")

        for line in state_to_be_removed.markers.values():
            line.remove()
        del self.states[state_to_be_removed.key]

    def remove_process(
        self, state_1: State, state_2: State, remove_states: bool = False
    ):
        """Remove a process from the self.process dictionary.

        The process to be removed is specified by the states that were used to
        initially create the process. It is optional to keep the points associated
        with the states while still removing the line object.

        Parameters
        ----------
        state_1: `~thermostate.thermostate.State`
            The starting state for this process.
        state_2: `~thermostate.thermostate.State`
            The final state for this process.
        remove_states: `bool`
            If ``True``, the associated states are removed from the instance.
        """
        key_1 = None
        key_2 = None
        for key, plotted_state in self.states.items():
            if state_1 is plotted_state.state:
                key_1 = key
            if state_2 is plotted_state.state:
                key_2 = key

        for line in self.processes[key_1 + key_2].values():
            line.remove()
        del self.processes[key_1 + key_2]

        if remove_states:
            self.remove_state(state_1)
            self.remove_state(state_2)

    def add_process(
        self,
        state_1: State,
        state_2: State,
        process_type: str | None = None,
        label_1: str | None = None,
        label_2: str | None = None,
    ):
        """Add a thermodynamic process to the self.process dictionary and plots it.

        A property of the states is held constant and all intermediate states are traced
        out in a line between the two states on the graph. The property that is held
        constant is specified by the user with the ``process_type`` input.
        If no property is to be held constant then a straight line between the
        two points is drawn.

        Parameters
        ----------
        state_1: `~thermostate.thermostate.State`
            The starting state for this process.
        state_2: `~thermostate.thermostate.State`
            The final state for this process.
        process_type: optional, `str`
            If given, specifies the property that is held constant during the process.
            Must be one of ``"isochoric"``, ``"isovolumetric"``, ``"isobaric"``,
            ``"isothermal"``, ``"isoenergetic"``, ``"isoenthalpic"``,
            ``"isentropic"``, or ``None``. If not specified, a straight line is drawn
            between the states.
        label_1: optional, `str`
            If given, will be used to label the first state.
        label_2: optional, `str`
            If given, will be used to label the second state.
        """
        if (
            process_type not in self.allowed_processes.keys()
            and process_type is not None
        ):
            raise ValueError(
                f"Not a supported process type: '{process_type}.\n"
                f"Supported process types are: {list(self.allowed_processes.keys())}"
            )

        if process_type is not None:
            constant_prop = self.allowed_processes[process_type]
            constant1 = getattr(state_1, constant_prop)
            constant2 = getattr(state_2, constant_prop)
            if not np.isclose(constant1, constant2):
                raise ValueError(f"Property: '{constant_prop}' was not held constant")

        missing_state_1 = True
        missing_state_2 = True
        key_1 = None
        key_2 = None
        sub1 = state_1.sub
        sub2 = state_2.sub
        if sub1 != sub2:
            raise ValueError(
                f"Substance of input states do not match: '{sub1}', '{sub2}'"
            )

        for key, plotted_state in self.states.items():
            if state_1 is plotted_state.state:
                missing_state_1 = False
                key_1 = key
            if state_2 is plotted_state.state:
                missing_state_2 = False
                key_2 = key

        if missing_state_1:
            key_1 = repr(state_1)
            self.add_state(state_1, key_1, label_1)

        if missing_state_2:
            key_2 = repr(state_2)
            self.add_state(state_2, key_2, label_2)

        plot_key = key_1 + key_2

        self.processes[plot_key] = {}

        if process_type in ("isochoric", "isovolumetric", "isometric"):
            p_1 = np.log10(state_1.p.magnitude)
            p_2 = np.log10(state_2.p.magnitude)
            v_range = np.logspace(p_1, p_2) * units.pascal
        elif process_type is not None:
            v_1 = np.log10(state_1.v.magnitude)
            v_2 = np.log10(state_2.v.magnitude)
            # Due to numerical approximation by CoolProp, an error occurs
            # if the state is too close to a saturated liquid. Here an
            # imperceptibly small offset is introduced to the specific volume
            # to avoid this error.
            if state_1.x is not None:
                if np.isclose(state_1.x.magnitude, 0.0):
                    v_1 *= 1.0 + 1.0e-14
                elif np.isclose(state_1.x.magnitude, 1.0):
                    v_1 *= 1.0 - 1.0e-12
            if state_2.x is not None:
                if np.isclose(state_2.x.magnitude, 0.0):
                    v_2 *= 1.0 + 1.0e-14
                elif np.isclose(state_2.x.magnitude, 1.0):
                    v_2 *= 1.0 - 1.0e-12
            v_range = np.logspace(v_1, v_2) * units("m**3/kg")

        for key, value in self.plots.items():
            x_data = []
            y_data = []
            fig, axis = value
            x_axis, y_axis = key

            if process_type is None:
                x_data.append(getattr(state_1, x_axis).magnitude)
                y_data.append(getattr(state_1, y_axis).magnitude)
                x_data.append(getattr(state_2, x_axis).magnitude)
                y_data.append(getattr(state_2, y_axis).magnitude)

                x_data = np.array(x_data) * getattr(units, self.axis_units[x_axis])
                y_data = np.array(y_data) * getattr(units, self.axis_units[y_axis])
                (line,) = axis.plot(x_data, y_data, marker="None", linestyle="--")
                self.processes[plot_key][key] = line
            else:
                state = State(state_1.sub)
                for v in v_range:
                    if process_type in ("isochoric", "isovolumetric", "isometric"):
                        state.pv = v, state_1.v
                    elif process_type == "isobaric":
                        state.pv = state_1.p, v
                    elif process_type == "isothermal":
                        state.Tv = state_1.T, v
                    elif process_type == "isoenergetic":
                        state.uv = state_1.u, v
                    elif process_type == "isoenthalpic":
                        state.hv = state_1.h, v
                    elif process_type == "isentropic":
                        state.sv = state_1.s, v

                    x_data.append(getattr(state, x_axis).magnitude)
                    y_data.append(getattr(state, y_axis).magnitude)

                x_data = np.array(x_data) * getattr(units, self.axis_units[x_axis])
                y_data = np.array(y_data) * getattr(units, self.axis_units[y_axis])
                (line,) = axis.plot(x_data, y_data, linestyle="-")
                self.processes[plot_key][key] = line

    def set_xscale(self, x_axis, y_axis, scale="linear"):
        """Access a plot in self.plots and change the scale of its x axis."""
        key = x_axis + y_axis
        fig, axis = self.plots[key]
        axis.set_xscale(scale)

    def set_yscale(self, x_axis, y_axis, scale="linear"):
        """Access a plot in self.plots and change the scale of its y axis."""
        key = x_axis + y_axis
        fig, axis = self.plots[key]
        axis.set_yscale(scale)


class VaporDome(PlottingBase):
    """Class for plotting graphs with a vapor dome."""

    def __init__(self, substance, *args):
        super().__init__(substance)
        min_temp = PropsSI("Tmin", substance)
        max_temp = PropsSI("Tcrit", substance)

        T_range = np.logspace(np.log10(min_temp), np.log10(max_temp), 400) * units.K
        self.st_f = [State(substance, T=T, x=0 * units.dimensionless) for T in T_range]
        self.st_g = [State(substance, T=T, x=1 * units.dimensionless) for T in T_range]
        for axes in args:
            self.plot(axes[0], axes[1])

    def plot(self, x_axis, y_axis):
        """Add a plot with a vapor dome to this instance with given x and y axes.

        Parameters
        ----------
        x_axis: `str`
            The string representing the x axis for this plot. Allowed axes are
            "T", "p", "u", "s", "v", and "h".
        y_axis: `str`
            The string representing the y axis for this plot. Allowed axes are
            "T", "p", "u", "s", "v", and "h".
        """
        if x_axis + y_axis not in self.plots:
            fig, axis = plt.subplots()
            self.plots[x_axis + y_axis] = (fig, axis)

            x_f = [getattr(st, x_axis).magnitude for st in self.st_f]
            x_f = np.array(x_f) * getattr(units, self.axis_units[x_axis])
            y_f = [getattr(st, y_axis).magnitude for st in self.st_f]
            y_f = np.array(y_f) * getattr(units, self.axis_units[y_axis])
            axis.plot(x_f, y_f)

            x_g = np.array(
                [getattr(st, x_axis).magnitude for st in self.st_g]
            ) * getattr(units, self.axis_units[x_axis])
            y_g = np.array(
                [getattr(st, y_axis).magnitude for st in self.st_g]
            ) * getattr(units, self.axis_units[y_axis])
            axis.plot(x_g, y_g)
            if x_axis in ("p", "v"):
                self.set_xscale(x_axis, y_axis, "log")
            if y_axis in ("p", "v"):
                self.set_yscale(x_axis, y_axis, "log")
        else:
            raise ValueError("Plot has already been added to this class instance")


class IdealGas(PlottingBase):
    """Class for plotting graphs modeled as an Ideal Gas."""

    def __init__(self, substance, *args):
        super().__init__(substance)
        for axes in args:
            self.plot(axes[0], axes[1])

    def plot(self, x_axis, y_axis):
        """Add a plot to this instance with given x and y axes.

        Parameters
        -----------
        x_axis: `str`
            The string representing the x axis for this plot. Allowed axes are
            "T", "p", "u", "s", "v", and "h".
        y_axis: `str`
            The string representing the y axis for this plot. Allowed axes are
            "T", "p", "u", "s", "v", and "h".
        """
        if x_axis + y_axis not in self.plots:
            fig, axis = plt.subplots()
            self.plots[x_axis + y_axis] = (fig, axis)
            if x_axis in ("p", "v"):
                self.set_xscale(x_axis, y_axis, "log")
            if y_axis in ("p", "v"):
                self.set_yscale(x_axis, y_axis, "log")
        else:
            raise ValueError("Plot has already been added to this class instance")



from .thermostate import units, Q_, State, set_default_units  # noqa
from .abbreviations import EnglishEngineering, SystemInternational  # noqa

try:
    import importlib.metadata as implib
except ImportError:
    import importlib_metadata as implib

__version__ = implib.version("thermostate")



"""Base ThermoState module."""
# Needed to support Python < 3.9
from __future__ import annotations
import sys
from typing import TYPE_CHECKING
import enum
from collections import OrderedDict

import CoolProp
from pint import UnitRegistry, DimensionalityError
from pint.unit import UnitsContainer, UnitDefinition
from pint.converters import ScaleConverter
import numpy as np

from .abbreviations import (
    SystemInternational as default_SI,
    EnglishEngineering as default_EE,
)

try:  # pragma: no cover
    from IPython.core.ultratb import AutoFormattedTB

except ImportError:  # pragma: no cover
    AutoFormattedTB = None

if TYPE_CHECKING:  # pragma: no cover
    import pint
    from typing import Union

units = UnitRegistry(autoconvert_offset_to_baseunit=True)
Q_ = units.Quantity
units.define(UnitDefinition("percent", "pct", (), ScaleConverter(1.0 / 100.0)))
units.setup_matplotlib()

default_units = None


def set_default_units(units):
    """Set default units to be used in class initialization."""
    if units is None or units in ("SI", "EE"):
        global default_units
        default_units = units
    else:
        raise TypeError(
            f"The given units '{units!r}' are not supported. Must be 'SI', "
            "'EE', or None."
        )


# Don't add the _render_traceback_ function to DimensionalityError if
# IPython isn't present. This function is only used by the IPython/ipykernel
# anyways, so it doesn't matter if it's missing if IPython isn't available.
if AutoFormattedTB is not None:  # pragma: no cover

    def render_traceback(self: DimensionalityError):
        """Render a minimized version of the DimensionalityError traceback.

        The default Jupyter/IPython traceback includes a lot of
        context from within pint that actually raises the
        DimensionalityError. This context isn't really needed for
        this particular error, since the problem is almost certainly in
        the user code. This function removes the additional context.
        """
        a = AutoFormattedTB(
            mode="Context", color_scheme="Neutral", tb_offset=1
        )  # type: ignore
        etype, evalue, tb = sys.exc_info()
        stb = a.structured_traceback(etype, evalue, tb, tb_offset=1)
        for i, line in enumerate(stb):
            if "site-packages" in line:
                first_line = slice(i)
                break
        else:
            # This is deliberately an "else" on the for loop
            first_line = slice(-1)
        return stb[first_line] + stb[-1:]

    DimensionalityError._render_traceback_ = render_traceback.__get__(  # type: ignore
        DimensionalityError
    )


class CoolPropPhaseNames(enum.Enum):
    """Map the phase names in CoolProp."""

    critical_point = CoolProp.iphase_critical_point
    gas = CoolProp.iphase_gas
    liquid = CoolProp.iphase_liquid
    not_imposed = CoolProp.iphase_not_imposed
    supercritical = CoolProp.iphase_supercritical
    supercritical_gas = CoolProp.iphase_supercritical_gas
    supercritical_liquid = CoolProp.iphase_supercritical_liquid
    twophase = CoolProp.iphase_twophase
    unknown = CoolProp.iphase_unknown


def munge_coolprop_input_prop(prop: str) -> str:
    """Munge an input property pair from CoolProp into our format.

    Example CoolProp input: ``XY_INPUTS``, where ``X`` and ``Y`` are one of
    ``T``, ``P``, ``Dmass``, ``Hmass``, ``Umass``, ``Q``, or ``Smass``. For
    use in ThermoState, we use lower case letters (except for T), replace
    ``D`` with ``v``, and replace ``Q`` with ``x``.

    Examples
    --------
    * ``DmassHmass_INPUTS``: ``vh``
    * ``DmassT_INPUTS``: ``vT``
    * ``PUmass_INPUTS``: ``pu``

    """
    prop = prop.replace("_INPUTS", "").replace("mass", "").replace("D", "V")
    return prop.replace("Q", "X").lower().replace("t", "T")


class StateError(Exception):
    """Errors associated with setting the `State` object."""

    def _render_traceback_(self):  # pragma: no cover
        """Render a minimized version of the `StateError` traceback.

        The default Jupyter/IPython traceback includes a lot of
        context from within `State` where the `StateError` is raised.
        This context isn't really needed, since the problem is almost certainly in
        the user code. This function removes the additional context.
        """
        if AutoFormattedTB is not None:
            a = AutoFormattedTB(mode="Context", color_scheme="Neutral", tb_offset=1)
            etype, evalue, tb = sys.exc_info()
            stb = a.structured_traceback(etype, evalue, tb, tb_offset=1)
            for i, line in enumerate(stb):
                if "site-packages" in line:
                    first_line = slice(i)
                    break
            else:
                first_line = slice(-1)
            return stb[first_line] + stb[-1:]


class State(object):
    """Basic State manager for thermodyanmic states.

    Parameters
    ----------
    substance : `str`
        One of the substances supported by CoolProp
    T : `pint.UnitRegistry.Quantity`
        Temperature
    p : `pint.UnitRegistry.Quantity`
        Pressure
    u : `pint.UnitRegistry.Quantity`
        Mass-specific internal energy
    s : `pint.UnitRegistry.Quantity`
        Mass-specific entropy
    v : `pint.UnitRegistry.Quantity`
        Mass-specific volume
    h : `pint.UnitRegistry.Quantity`
        Mass-specific enthalpy
    x : `pint.UnitRegistry.Quantity`
        Quality

    """

    _allowed_subs = [
        "AIR",
        "AMMONIA",
        "WATER",
        "PROPANE",
        "R134A",
        "R22",
        "ISOBUTANE",
        "CARBONDIOXIDE",
        "OXYGEN",
        "NITROGEN",
    ]

    _all_pairs = set(
        munge_coolprop_input_prop(k)
        for k in dir(CoolProp.constants)
        if "INPUTS" in k and "molar" not in k
    )
    _all_pairs.update([k[::-1] for k in _all_pairs])

    _unsupported_pairs = {"Tu", "Th", "us", "hx"}
    _unsupported_pairs.update([k[::-1] for k in _unsupported_pairs])

    _allowed_pairs = _all_pairs - _unsupported_pairs

    _all_props = set("Tpvuhsx")

    _read_only_props = {"cp", "cv", "phase"}

    _dimensions = {
        "T": UnitsContainer({"[temperature]": 1.0}),
        "p": UnitsContainer({"[mass]": 1.0, "[length]": -1.0, "[time]": -2.0}),
        "v": UnitsContainer({"[length]": 3.0, "[mass]": -1.0}),
        "u": UnitsContainer({"[length]": 2.0, "[time]": -2.0}),
        "h": UnitsContainer({"[length]": 2.0, "[time]": -2.0}),
        "s": UnitsContainer({"[length]": 2.0, "[time]": -2.0, "[temperature]": -1.0}),
        "x": UnitsContainer({}),
    }

    _SI_units = {
        "T": "kelvin",
        "p": "pascal",
        "v": "meter**3/kilogram",
        "u": "joules/kilogram",
        "h": "joules/kilogram",
        "s": "joules/(kilogram*kelvin)",
        "x": "dimensionless",
        "cp": "joules/(kilogram*kelvin)",
        "cv": "joules/(kilogram*kelvin)",
    }

    def __setattr__(
        self,
        key: str,
        value: "Union[str, pint.Quantity, tuple[pint.Quantity, pint.Quantity]]",
    ) -> None:
        if key.startswith("_") or key in ("sub", "label", "units"):
            object.__setattr__(self, key, value)
        elif key in self._allowed_pairs:
            if not isinstance(value, tuple):  # pragma: no cover, for typing
                raise ValueError("Must pass a tuple of Quantities")
            self._check_dimensions(key, value)
            self._check_values(key, value)
            self._set_properties(key, value)
        elif key in self._unsupported_pairs:
            raise StateError(
                f"The pair of input properties entered ({key}) isn't supported yet. "
                "Sorry!"
            )
        else:
            raise AttributeError(f"Unknown attribute {key}")

    def __getattr__(
        self, key: str
    ) -> "Union[str, tuple[pint.Quantity, pint.Quantity], pint.Quantity]":
        if key in self._all_props:
            return object.__getattribute__(self, "_" + key)
        elif key in self._all_pairs:
            val_0 = object.__getattribute__(self, "_" + key[0])
            val_1 = object.__getattribute__(self, "_" + key[1])
            return val_0, val_1
        elif key == "phase":
            return object.__getattribute__(self, "_" + key)
        elif key in self._read_only_props:
            return object.__getattribute__(self, "_" + key)
        else:
            raise AttributeError(f"Unknown attribute {key}")

    def __eq__(self, other: object) -> bool:
        """Check if two `State`s are equivalent.

        Check that they are using the same substance and two properties that are
        always independent. Choose T and v because the EOS tends to be defined
        in terms of T and density.
        """
        if not isinstance(other, State):
            return NotImplemented
        if (
            self.sub == other.sub
            # Pylance does not support NumPy ufuncs
            and np.isclose(other.T, self.T)  # type: ignore
            and np.isclose(other.v, self.v)  # type: ignore
        ):
            return True
        return False

    def __le__(self, other: "State"):
        return NotImplemented

    def __lt__(self, other: "State"):
        return NotImplemented

    def __gt__(self, other: "State"):
        return NotImplemented

    def __ge__(self, other: "State"):
        return NotImplemented

    def __init__(
        self, substance: str, label=None, units=None, **kwargs: "pint.Quantity"
    ):

        if units is None:
            units = default_units
        self.units = units

        self.label = label

        if substance.upper() in self._allowed_subs:
            self.sub = substance.upper()
        else:
            raise ValueError(
                f"{substance} is not an allowed substance. "
                f"Choose one of {self._allowed_subs}."
            )

        self._abstract_state = CoolProp.AbstractState("HEOS", self.sub)

        input_props = ""
        for arg in kwargs:
            if arg not in self._all_props:
                raise ValueError(f"The argument {arg} is not allowed.")
            else:
                input_props += arg

        if len(input_props) > 2 or len(input_props) == 1:
            raise ValueError(
                "Incorrect number of properties specified. Must be 2 or 0."
            )

        if len(input_props) > 0 and input_props not in self._allowed_pairs:
            raise StateError(
                f"The pair of input properties entered ({input_props}) isn't supported "
                "yet. Sorry!"
            )

        if len(input_props) > 0:
            setattr(self, input_props, (kwargs[input_props[0]], kwargs[input_props[1]]))

    @property
    def label(self):
        """Get or set the string label for this state, used in plotting."""
        return self._label

    @label.setter
    def label(self, value: str | None):
        if value is None:
            self._label = value
            return
        try:
            label = str(value)
        except Exception:
            raise TypeError(
                f"The given label '{value!r}' could not be converted to a string"
            ) from None
        self._label = label

    @property
    def units(self):
        """Get or set the string units for this state to set attribute units."""
        return self._units

    @units.setter
    def units(self, value: str | None):
        if value is None or value in ("EE", "SI"):
            self._units = value
            if hasattr(self, "T"):
                setattr(self, "Tv", (self.T, self.v))
        else:
            raise TypeError(
                f"The given units '{units!r}' are not supported. Must be 'SI', "
                "'EE', or None."
            )

    def to_SI(self, prop: str, value: "pint.Quantity") -> "pint.Quantity":
        """Convert the input ``value`` to the appropriate SI base units."""
        return value.to(self._SI_units[prop])

    def to_PropsSI(self, prop: str, value: "pint.Quantity") -> float:  # noqa: D403
        """CoolProp can't handle Pint Quantites so return the magnitude only.

        Convert to the appropriate SI units first.
        """  # noqa: D403
        return self.to_SI(prop, value).magnitude

    @staticmethod
    def _check_values(
        properties: str, values: "tuple[pint.Quantity, pint.Quantity]"
    ) -> None:
        for p, v in zip(properties, values):
            if p in "Tvp" and v.to_base_units().magnitude < 0.0:
                raise StateError(f"The value of {p} must be positive in absolute units")
            elif p == "x" and not (0.0 <= v.to_base_units().magnitude <= 1.0):
                raise StateError("The value of the quality must be between 0 and 1")

    def _check_dimensions(
        self, properties: str, values: "tuple[pint.Quantity, pint.Quantity]"
    ) -> None:
        for p, v in zip(properties, values):
            # Dimensionless values are a special case and don't work with
            # the "check" method.
            try:
                valid = v.check(self._SI_units[p])
            except KeyError:
                valid = v.dimensionality == self._dimensions[p]
            if not valid:
                raise StateError(
                    f"The dimensions for {p} must be {self._dimensions[p]}"
                )

    def _set_properties(
        self, known_props: str, known_values: "tuple[pint.Quantity, pint.Quantity]"
    ) -> None:
        known_state: OrderedDict[str, float] = OrderedDict()

        for prop, val in zip(known_props, known_values):
            if prop == "x":
                known_state["Q"] = self.to_PropsSI(prop, val)
            elif prop == "v":
                known_state["Dmass"] = 1.0 / self.to_PropsSI(prop, val)
            else:
                postfix = "" if prop in ["T", "p"] else "mass"
                known_state[prop.upper() + postfix] = self.to_PropsSI(prop, val)

        for key in sorted(known_state):
            known_state.move_to_end(key)

        inputs = getattr(CoolProp, "".join(known_state.keys()) + "_INPUTS")
        try:
            self._abstract_state.update(inputs, *known_state.values())
        except ValueError as e:
            if "Saturation pressure" in str(e):
                raise StateError(
                    f"The given values for {known_props[0]} and {known_props[1]} are "
                    "not independent."
                )
            else:
                raise

        for prop in self._all_props.union(self._read_only_props):
            if prop == "v":
                v = 1.0 / self._abstract_state.keyed_output(CoolProp.iDmass)
                value = v * units(self._SI_units[prop])
            elif prop == "x":
                x = self._abstract_state.keyed_output(CoolProp.iQ)
                if x == -1.0:
                    value = None
                else:
                    value = x * units(self._SI_units[prop])
            elif prop == "phase":
                value = CoolPropPhaseNames(
                    self._abstract_state.keyed_output(CoolProp.iPhase)
                ).name
            else:
                postfix = "" if prop in "Tp" else "mass"
                p = getattr(CoolProp, "i" + prop.title() + postfix)
                value = self._abstract_state.keyed_output(p) * units(
                    self._SI_units[prop]
                )

            set_units = None
            if self.units == "SI":
                set_units = getattr(default_SI, prop, None)
            elif self.units == "EE":
                set_units = getattr(default_EE, prop, None)
            if set_units is not None:
                value.ito(set_units)
            setattr(self, "_" + prop, value)



"""
This module contains classes with attributes representing the common property units.

Example
-------
These classes are shortcuts to the units for common properties::

    >>> st = State("water", T=Q_(300.0, "K"), p=Q_(101325, "Pa"))
    >>> h = st.h.to(SI.h)
    >>> u = st.u.to(EE.u)

"""


class EnglishEngineering:
    """String representations of common units.

    The attributes of this class are strings that represent the common units for
    thermodynamics calculations.

    Attributes
    ----------
    h : `str`
        BTU/lb
    p : `str`
        psi
    s : `str`
        BTU/(lb*degR)
    T : `str`
        degF
    u : `str`
        BTU/lb
    v : `str`
        ft**3/lb
    cp : `str`
        BTU/(lb*degR)
    cv : `str`
        BTU/(lb*degR)

    """

    s = "BTU/(lb*degR)"
    h = "BTU/lb"
    T = "degF"
    u = "BTU/lb"
    v = "ft**3/lb"
    p = "psi"
    cp = "BTU/(lb*degR)"
    cv = "BTU/(lb*degR)"


class SystemInternational:
    """String representations of common units.

    The attributes of this class are strings that represent the common units for
    thermodynamics calculations.

    Attributes
    ----------
    h : `str`
        kJ/kg
    p : `str`
        bar
    s : `str`
        kJ/(kg*K)
    T : `str`
        degC
    u : `str`
        kJ/kg
    v : `str`
        m**3/lb
    cv : `str`
        kJ/(K*kg)
    cp : `str`
        kJ/(K*kg)

    """

    s = "kJ/(kg*K)"
    h = "kJ/kg"
    T = "degC"
    u = "kJ/kg"
    v = "m**3/kg"
    p = "bar"
    cv = "kJ/(K*kg)"
    cp = "kJ/(K*kg)"
